import asyncio
import inspect

def run(*args):

    coros = list(filter(inspect.iscoroutine, args))
    future = None

    if len(coros) == 1:
        future = coros[0]
    elif len(coros) > 1:
        future = asyncio.gather(*coros)

    if future:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)



class WebException(Exception):
    """Web request returns unsucessful status."""

    def __init__(self, status, text, msg = None):
        self.status = status
        self.text = text
        msg = msg or "Got status code other than 200."
        super().__init__(msg)

class NotFound(WebException):
    """Web request returned 404 status code."""

    def __init__(self, text):
        msg = "Got status code 404."
        super().__init__(404, text, msg)



from .ws import MixerWS

class MixerConstellation:

    CONSTELLATION_URL = "wss://constellation.mixer.com"
    websocket = None

    def __init__(self, on_connected):
        self.on_connected = on_connected
        self.callbacks = dict()
        self.packet_id = 0

    async def start(self):
        """Initializes the Constellation websocket and begins to listen for events."""

        self.websocket = MixerWS(self.CONSTELLATION_URL)
        await self.websocket.connect()
        await self.on_connected(self) # call on_connected func (we should probably subscribe to events)

        while True:

            # receive a packet from server
            packet = await self.websocket.receive_packet()

            # make sure it's an event we're subscribed to
            if packet["type"] != "event": continue
            if packet["event"] != "live": continue

            # find and invoke the callback function with the packet & payload
            event_name = packet["data"]["channel"]
            payload = packet["data"]["payload"]
            callback = self.callbacks.get(event_name, None)
            if callback is not None:
                await callback(packet, payload)

    async def subscribe(self, events, callback):
        """Subcribes the Constellation websocket to a list of provided events."""

        # if a single event is provided, wrap it in a list automatically
        if isinstance(events, str):
            events = [events]

        # build livesubscribe packet
        packet = {
            "type": "method",
            "method": "livesubscribe",
            "params": {
                "events": events
            },
            "id": self.packet_id
        }

        # send packet to server and determine callback
        await self.websocket.send_packet(packet)
        for event in events:
            self.callbacks[event] = callback

        # increment packet id and return unique packet id
        self.packet_id += 1
        return self.packet_id



import json
import websockets
import inspect

class MixerWS():

    def __init__(self, url, **kwargs):
        self.url = url
        self.on_connected = kwargs.pop("on_connected", None)
        self.kwargs = kwargs

    async def try_call(self, func, *opts):
        """Calls a coroutine function with parameters, if it's defined."""
        if inspect.iscoroutinefunction(func):
            await func(*opts)

    async def connect(self):
        """Establishes connection to websocket endpoint and calls on_connected callback."""
        self.websocket = await websockets.connect(self.url, **self.kwargs)
        await self.try_call(self.on_connected)

    async def send_packet(self, packet):
        """Sends a packet to the server.

        Args:
            packet (dict): Data to be json encoded and send.
        """
        packet_raw = json.dumps(packet)
        await self.websocket.send(packet_raw)

    async def receive_packet(self):
        """dict: Receives a packet from the server."""
        packet_raw = await self.websocket.recv()
        return json.loads(packet_raw)



import aiohttp
import dateutil.parser
import json
from datetime import datetime, timezone, timedelta
from enum import Enum

from . import exceptions as MixerExceptions
from .objects import MixerUser, MixerChannel

class RequestMethod(Enum):
    GET = 0
    POST = 1

class MixerAPI:

    API_URL = "https://mixer.com/api/v1"
    API_URL_V2 = "https://mixer.com/api/v2"

    def __init__(self, client_id, client_secret):
        self.client_id = client_id
        self.client_secret = client_secret
        self._session = aiohttp.ClientSession(headers = { "Client-ID": self.client_id })

    async def close(self):
        await self._session.close()

    async def request(self, method, url, parse_json = False, **kwargs):

        # pick ... based on request type
        if method is RequestMethod.GET:
            ctx_mgr = self._session.get(url, **kwargs)
        elif method is RequestMethod.POST:
            data = kwargs.pop("data")
            ctx_mgr = self._session.post(url, json = data, **kwargs)

        async with ctx_mgr as response:

            text = await response.text()

            # handle specific response codes
            if response.status == 404:
                raise MixerExceptions.NotFound(text)

            if response.status != 200:
                raise MixerExceptions.WebException(response.status, text)

            try:
                # NOTE: the only exception that will be thrown here is in the .json() func
                return await response.json() if parse_json else text
            except:
                raise RuntimeError("Failed to parse json from response.")

    async def get(self, url, **kwargs):
        coro = self.request(RequestMethod.GET, url, **kwargs)
        return await coro

    async def post(self, url, data, **kwargs):
        kwargs["data"] = data
        coro = self.request(RequestMethod.POST, url, **kwargs)
        return await coro

    async def get_channel(self, id_or_token):
        """Retrieves a MixerChannel object from username or channel id.

        Args:
            id_or_token (str): Username (or id) of Mixer channel.

        Returns:
            :class:`mixer.objects.MixerChannel`: Channel information.
        """
        url = "{}/channels/{}".format(self.API_URL, id_or_token)
        data = await self.get(url, parse_json = True)
        channel = MixerChannel(data)
        channel.set_api(self)
        return channel

    async def get_user(self, user_id):
        """Retrieves a MixerUser object from a user id.

        Args:
            user_id (int): The unique id of a Mixer user.

        Returns:
            :class:`mixer.objects.MixerUser`: User information.
        """
        url = "{}/users/{}".format(self.API_URL, user_id)
        data = await self.get(url, parse_json = True)
        user = MixerUser(data)
        user.set_api(self)
        return user

    async def get_shortcode(self, scope = None):
        """Makes a request to begin shortcode oauth process.

        Args:
            scope (list): A list of scope/permissions to generate token with.

        Returns:
            dict: Information to proceed with shortcode oauth process."""
        url = "{}/oauth/shortcode".format(self.API_URL)
        if scope is None: scope = list()
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "scope": " ".join(scope)
        }
        response = await self.post(url, data, parse_json = True)
        return response

    async def check_shortcode(self, handle):
        """Check a shortcode handle to determine it's status.

        Args:
            str: Shortcode handle.

        Returns:
            dict: Shortcode status information.
        """
        url = "{}/oauth/shortcode/check/{}".format(self.API_URL, handle)
        response = await self.get(url, parse_json = True)
        return response

    async def get_token(self, code_or_token, refresh = False):
        """Generate/refresh tokens.

        Args:
            code_or_token (str): Authorization code or refresh token.
            refresh (bool): Whether or not a refresh token is provided.

        Returns:
            dict: New token(s) + information from server.
        """
        url = "{}/oauth/token".format(self.API_URL)
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret
        }

        if refresh:
            data["grant_type"] = "refresh_token"
            data["refresh_token"] = code_or_token
        else:
            data["grant_type"] = "authorization_code"
            data["code"] = code_or_token

        response = await self.post(url, data, parse_json = True)
        return response # https://pastebin.com/n1Kjjphq

    async def check_token(self, token):
        """Gets information about an existing token.

        Args:
            token (str): An access token or a refresh token.
        """
        url = "{}/oauth/token/introspect".format(self.API_URL)
        data = { "token": token }
        response = await self.post(url, data, parse_json = True)
        return response # https://pastebin.com/SEd6Y2Jz

    async def get_broadcast(self, channel_id):
        """Gets an active broadcast on a given chanel.

        Args:
            channel_id (int): Unique channel ID number.
        """
        url = "{}/channels/{}/broadcast".format(self.API_URL, channel_id)
        response = await self.get(url, parse_json = True)
        return response

    async def get_uptime(self, channel_id):
        """Gets the uptime of a channels broadcast.

        Returns:
            datetime.timedelta: Duration of the active broadcast.
            None: If the broadcast is not currently online.
        """

        # get broadcast and make sure it's online
        broadcast = await self.get_broadcast(channel_id)
        if "error" in broadcast or not broadcast["online"]:
            return None

        # determine the streams start time and current time
        started = dateutil.parser.parse(broadcast["startedAt"])
        now = datetime.now(timezone.utc)

        # calculate delta and remove microseconds because they're insignificant
        delta = now - started
        delta = delta - timedelta(microseconds = delta.microseconds)
        return delta

    async def get_leaderboard(self, type, channel_id, limit = 10):
        # type format: [sparks, embers]-[weekly, monthly, yearly, alltime]
        url = "{}/leaderboards/{}/channels/{}?limit={}".format(self.API_URL_V2, type, channel_id, limit)
        response = await self.get(url, parse_json = True)
        return response

    async def get_chatters(self, channel_id):
        url = "{}/chats/{}/users".format(self.API_URL_V2, channel_id)
        response = await self.get(url, parse_json = True)
        return response

    async def get_user_services(self, oauth):
        # NOTE: requires "user:details:self" scope
        await oauth.ensure_active()
        url = "{}/users/{}/links".format(self.API_URL, oauth.user_id)
        response = await self.get(url, parse_json = True, headers = oauth.header)
        return response

    async def get_user_service(self, service, oauth):
        # NOTE: requires "user:details:self" scope
        services = await self.get_user_services(oauth)
        check = lambda d: d["service"] == service
        filtered = list(filter(check, services))
        return filtered[0]["id"] if len(filtered) else None



# python setup.py sdist bdist_wheel
# twine upload --repository-url https://test.pypi.org/legacy/ dist/*
# pip install --index-url https://test.pypi.org/simple/ --no-deps mixer.py

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mixer.py",
    version="0.0.2",
    author="Justin Garofolo",
    author_email="justin@garofolo.net",
    description="An unofficial Mixer API wrapper written in Python. ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ooojustin/mixer.py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import json, asyncio, inspect, logging
from time import time

from . import exceptions as MixerExceptions

# mixer recaptcha v2 sitekey:
# 6LeYS2gUAAAAAPVr3SzjSJYtfD7iBxS5yyWS0IuH

# https://mixer.com/api/v1/users/login
# post: username, password, captcha
# create session and automate oauth login?

class MixerOAuth:

    @classmethod
    async def create_from_authorization_code(cls, api, code):
        self = MixerOAuth()
        self.api = api
        self.refresh_token = code
        await self.refresh(is_refresh = False)
        return self


    @classmethod
    async def create(cls, api, access_token, refresh_token):
        self = MixerOAuth()
        self.api = api
        self.access_token = access_token
        self.refresh_token = refresh_token
        await self.update_token_data()
        return self

    def __init__(self):

        # array of methods to invoke once access_token is refreshed
        # methods are invoked with 2 params (new access_token/refresh_token)
        self._refreshed = list()

        # scheduled task to refresh tokens
        self._auto_refresh_task = None

    async def update_token_data(self):
        data = await self.api.check_token(self.access_token)
        active = data.get("active")
        self.expires = data.get("exp") - 10 if active else -1
        self.user_id = data.get("sub") if active else -1
        self.username = data.get("username") if active else ""

    async def ensure_active(self):
        """Refreshes the access_token if it's not active."""
        if not self.active:
            await self.refresh()

    async def refresh(self, auto_refreshed = False, is_refresh = True):
        """Refreshes tokens and triggers callbacks w/ new tokens.

        Args:
            auto_refreshed (bool): Indicates that this token refresh was automated by a task.
            is_refresh (bool): If false, the refresh token will be used as an authorization code.
        """

        # refresh tokens
        tokens = await self.api.get_token(self.refresh_token, refresh = is_refresh)
        self.access_token = tokens.get("access_token")
        self.refresh_token = tokens.get("refresh_token")
        await self.update_token_data()

        # trigger callbacks w/ new tokens
        for event in self._refreshed:
            if inspect.iscoroutinefunction(event):
                await event(self.access_token, self.refresh_token)
            elif inspect.isfunction(event):
                event(self.access_token, self.refresh_token)

        # if this wasn't trigerred by auto-refresh
        # and we have an auto-refresh task running
        # automatically re-register the task
        if (not auto_refreshed
                and self._auto_refresh_task is not None
                and not self._auto_refresh_task.cancelled()
                and not self._auto_refresh_task.done()):
            self._auto_refresh_task.cancel()
            self.register_auto_refresh()

    async def _auto_refresh(self):

        # determine time to wait before a refresh, make sure it's at least 0
        delay = self.expires - time()
        if delay < 0:
            delay = 0

        # sleep until we should refresh the token
        print("waiting {} seconds before automatically refreshing tokens...".format(delay))
        await asyncio.sleep(delay)

        # refresh tokens
        await self.refresh(auto_refreshed = True)

        # schedule this function as a task so we can automatically do it again
        self.register_auto_refresh()

    def register_auto_refresh(self):
        """Registers a task that will endlessly ensure the tokens are valid."""
        self._auto_refresh_task = asyncio.create_task(self._auto_refresh())

    def on_refresh(self, callback):
        """Adds a callback to be triggered when tokens are updated."""
        self._refreshed.append(callback)

    @property
    def header(self):
        """dict: A dictionary containing the oauth bearer token."""
        return { "Authorization": "Bearer {}".format(self.access_token) }

    @property
    def active(self):
        """bool: Indicates if the access_token is still valid."""
        return self.expires > time()



import dateutil.parser

# https://dev.mixer.com/rest/index.html#TimeStamped
class TimeStamped:

    def __init__(self, data):
        self.data = data

    def __datetime(self, name):
        str = self.data.get(name)
        return dateutil.parser.parse(str) if str else None

    @property
    def created_at(self):
        """datetime: The date/time the data was created."""
        return self.__datetime("createdAt")

    @property
    def updated_at(self):
        """datetime: The date/time the data was updated."""
        return self.__datetime("updatedAt")

    @property
    def deleted_at(self):
        """datetime: The date/time the data was deleted."""
        return self.__datetime("deletedAt")

# https://dev.mixer.com/rest/index.html#UserWithChannel
class MixerUser(TimeStamped):

    def __init__(self, data, channel = None):
        super().__init__(data)
        self.data = data

        # determine channel information
        if isinstance(channel, MixerChannel):
            self._channel = channel
        else:
            channel_data = self.data.get("channel")
            self._channel = MixerChannel(channel_data, self)

    api = None
    def set_api(self, api):
        self.api = api
        self.channel.api = api

    @property
    def avatar_url(self):
        """str: A link to the users avatar on Mixer."""
        return self.data.get("avatarUrl")

    @property
    def bio(self):
        """str: The users biography on their Mixer profile. This may contain HTML."""
        return self.data.get("bio")

    @property
    def channel(self):
        """:class:`mixer.objects.MixerChannel`: Information about the Mixer channel associated with this user."""
        return self._channel

    @property
    def experience(self):
        """int: The user's experience points."""
        return self.data.get("experience")

    @property
    def groups(self):
        """list: A list of groups that the user is in. Each contains 'id' and 'name'."""
        return self.data.get("groups")

    @property
    def id(self):
        """int: The unique ID of the user."""
        return self.data.get("id")

    @property
    def level(self):
        """int: The user's current level on Mixer, as determined by the number of experience points the user has."""
        return self.data.get("level")

    @property
    def social(self):
        """dict: Social links."""
        return self.data.get("social")

    @property
    def sparks(self):
        """int: The amount of sparks the user has."""
        return self.data.get("sparks")

    @property
    def username(self):
        """str: The user's name. This is unique on the site and is also their channel name."""
        return self.data.get("username")

    @property
    def verified(self):
        """bool: Indicates whether the user has verified their e-mail."""
        return self.data.get("verified")

# https://dev.mixer.com/rest/index.html#ExpandedChannel
class MixerChannel:

    def __init__(self, data, user = None):
        self.data = data

        # determine user information
        if isinstance(user, MixerUser):
            self._user = user
        else:
            user_data = self.data.get("user")
            self._user = MixerUser(user_data, self)

    api = None
    def set_api(self, api):
        self.api = api
        self.user.api = api

    @property
    def id(self):
        """int: The unique ID of the channel."""
        return self.data.get("id")

    @property
    def username(self):
        """str: The name and url of the channel."""
        return self.data.get("token")

    @property
    def online(self):
        """bool: Indicates if the channel is currently streaming."""
        return self.data.get("online")

    @property
    def user(self):
        """:class:`mixer.objects.MixerUser`: Information about the Mixer user associated with this channel."""
        return self._user

    @property
    def featured(self):
        """bool: True if feature_level is > 0."""
        return self.data.get("featured")

    @property
    def feature_level(self):
        """int: The featured level for this channel. Its value controls the position and order of channels in the featured carousel."""
        return self.data.get("featureLevel")

    @property
    def partnered(self):
        """bool: Indicates if the channel is partnered."""
        return self.data.get("partnered")

    @property
    def transcoding_profile_id(self):
        """int: The id of the transcoding profile."""
        return self.data.get("transcodingProfileId")

    @property
    def suspended(self):
        """bool: Indicates if the channel is suspended."""
        return self.data.get("suspendeded")

    @property
    def name(self):
        """str: The title of the channel."""
        return self.data.get("name")

    @property
    def audience(self):
        """str ('family', 'teen', '18+'): The target audience of the channel."""
        return self.data.get("audience")

    @property
    def viewers_total(self):
        """int: Amount of unique viewers that have ever viewed this channel."""
        return self.data.get("viewersTotal")

    @property
    def viewers(self):
        """int: Amount of viewers currently watching this channel."""
        return self.data.get("viewersCurrent")

    @property
    def followers(self):
        """int: Amount of people following this channel."""
        return self.data.get("numFollowers")

    @property
    def description(self):
        """str: The description of the channel. Note that this may contain HTML."""
        return self.data.get("description")

    @property
    def game_type(self):
        """dict: The type of game broadcasted on the users channel."""
        return self.data.get("type")

    @property
    def interactive(self):
        """bool: Indicates if the channel is interactive."""
        return self.data.get("interactive")

    @property
    def interactive_game_id(self):
        """int: The ID of the interactive game used."""
        return self.data.get("interactiveGameId")

    @property
    def ftl(self):
        """int: The FTL stream id."""
        return self.data.get("ftl")

    @property
    def has_vod(self):
        """bool: Indicates if the channel has vod saved."""
        return self.data.get("hasVod")

    @property
    def language_id(self):
        """str: The ISO 639 language id."""
        return self.data.get("languageId")

    @property
    def banner_url(self):
        """str: The URL of the banner image."""
        return self.data.get("bannerUrl")

    @property
    def hosteeId(self):
        """int: The ID of the hostee channel."""
        return self.data.get("hosteeId")

    @property
    def has_transcodes(self):
        """bool: Indicates if the channel has transcodes enabled."""
        return self.data.get("hasTranscodes")

    @property
    def vods_enabled(self):
        """bool: Indicates if the channel has vod recording enabled."""
        return self.data.get("vodsEnabled")

    @property
    def costream_id(self):
        """str: The costream that the channel is in, if any."""
        return self.data.get("costreamId")

    @property
    def thumbnail(self):
        """dict: The resource information regarding the channel thumbnail."""
        return self.data.get("thumbnail")

    @property
    def cover(self):
        """dict: The resource information regarding the channel cover."""
        return self.data.get("cover")

    @property
    def badge(self):
        """dict: The resource information regarding the channel badge for subscribers."""
        return self.data.get("badge")

    async def get_leaderboard(self, type, limit = 10):
        """dict: Gets a list of users on a specified leaderboard."""
        return await self.api.get_leaderboard(type, self.id, limit)

    async def get_uptime(self):
        """datetime.timedelta: The duration of the active broadcast."""
        return await self.api.get_uptime(self.id)

# https://pastebin.com/NW6NcS8z
class MixerChatMessage:

    def __init__(self, data):
        self.data = data

    @property
    def id(self):
        """str: The unique identifier of the message."""
        return self.data.get("id")

    @property
    def username(self):
        """str: The username of the person who sent the message."""
        return self.data.get("user_name")

    @property
    def user_id(self):
        """int: The id of the user who sent the message."""
        return self.data.get("user_id")

    @property
    def user_avatar(self):
        """str: The avatar of the user who sent the message."""
        return self.data.get("user_avatar")

    @property
    def roles(self):
        """list: The roles the user has in the chat room."""
        return self.data.get("user_roles")

    @property
    def message_raw(self):
        """dict: The raw message data retrieved from the server."""
        return self.data.get("message")

    def has_role(self, role):
        """Determines whether or not the sender of the message has a specified role.

        Args:
            role (str): The role to check the sender for. (ex: 'Owner')

        Returns:
            bool: Indicates whether or not the message sender has the role.
        """
        return role in self.roles

    @property
    def text(self):
        """str: The raw text of the message."""
        text = ""
        for piece in self.message_raw["message"]:
            text += piece["text"]
        return text

    @property
    def tags(self):
        """list: A list of usernames tagged in this message, in order."""
        tags = list()
        for piece in self.message_raw["message"]:
            if piece["type"] == "tag":
                tags.append(piece["username"])
        return tags

    @property
    def skill(self):
        """dict: Gets the skill used that's associated with the message."""
        return self.message_raw["meta"].get("skill")

    @property
    def is_skill(self):
        """bool: Indicates if the message is a skill."""
        return self.message_raw["meta"].get("is_skill", False)

    @property
    def is_whisper(self):
        """bool: Indicates if the message is a whisper."""
        return self.message_raw["meta"].get("whisper", False)

    @property
    def is_censored(self):
        """bool: Indicates if the message is censored by Catbot's auto-moderation."""
        return self.message_raw["meta"].get("whisper", False)

    async def delete(self):
        """Deletes this message from the chat."""
        await self.chat.send_method_packet("deleteMessage", self.id)



import inspect
import requests
import shlex
import asyncio
from enum import Enum

from .ws import MixerWS
from .objects import MixerChatMessage

class MixerChat:

    class ParamType(Enum):
        NUMBER = 0,
        POSITIVE_NUMBER = 1,
        MIXER_USER = 2

    class ChatCommands:

        commands = dict()

        def get_command(self, name, param_count = None):
            """Gets a chat command from the name and number of parameters.

            Args:
                name (str): The name of the command.
                param_count (int): The number of parameters expected.
            """

            # make sure the command actually exists
            if not name in self.commands:
                return None

            # get a list of overloaded commands
            command_list = self.commands[name]

            # if parma_count isnt specified, return the first defined func
            if param_count is None:
                return command_list[0]

            # look for a definition with a matching parameter count
            for command in command_list:
                if command["param_count"] == param_count:
                    return command

            # return false because the command is defined but no matching parameter count
            return False

        def get_help(self, name, param_count = None):
            """Gets a description of a specific command.

            Args:
                name (str): The name of the command.
                param_count (int): The number of parameters expected.

            Returns:
                str: A description/documentation of a chat command.
            """

            command = self.get_command(name, param_count)

            if command is False:
                return "failed to find '{}' command with {} parameters.".format(name, param_count)
            elif command is None:
                return "failed to find '{}' command.".format(name)
            elif command["description"] is None:
                return "command '{}' with {} parameters is undocumented.".format(name, param_count)

            if command["param_count"] > 0:
                params = ", ".join(command["params"])
                return "{} ({}) -> {}".format(name, params, command["description"])
            else:
                return "{} -> {}".format(name, command["description"])

        def add_command(self, name, func):
            """Manually adds a new command, given a name a reference to the callable.

            Args:
                name (str): Name of the command.
                func (function): The function to link this command to.
            """

            if not inspect.iscoroutinefunction(func):
                return

            name = name.lower()
            sig = inspect.signature(func)
            params = sig.parameters
            command = {
                "function": func,
                "signature": sig,
                "description": func.__doc__, # command docstring (should be a brief description)
                "params": list(params.keys())[1:], # list of parameter names
                "param_count": len(params) - 1 # ignore data parameter (required)
            }

            existing = self.get_command(name, command["param_count"])
            if isinstance(existing, dict):
                # this command name already exists with this parameter count
                # it cant be overloaded unless we override the old one
                return

            if name in self.commands:
                self.commands[name].append(command)
            else:
                self.commands[name] = [command]

        async def trigger_command(self, command, message, params):
            response = await command["function"](message, *params)
            if response is not None:
                response = "@{} {}".format(message.username, response)
                await self.chat.send_message(response)

        async def handle(self, message):
            """Handle/parse a chat message as a command.

            Args:
                message (MixerChatMessage): A chat message wrapper.

            Returns:
                bool: Indicates if the message was handled as a command.
            """

            # command prefix check
            pl = len(self.prefix)
            if message.text[:pl] != self.prefix:
                return False

            # handle it as a command
            try:
                parsed = shlex.split(message.text) # split string by whitespace and account for quotes
                name = parsed[0][pl:].lower() # the name of the command -> 0th item with command prefix removed
                parameters = parsed[1:] # remove first parsed item, because its the command name
            except:
                await self.chat.send_message("an error occurred while parsing that command.")
                return True

            # make sure the command exists
            command = self.get_command(name, len(parameters))
            if command is None:
                await self.chat.send_message("unrecognized command '{}'.".format(name))
                return True
            elif command is False:
                await self.chat.send_message("invalid parameter count for command '{}'.".format(name))
                return True

            # handle parameter type annotations
            ParamType = self.chat.ParamType
            param_names = command["params"]
            param_objects = command["signature"].parameters
            for i in range(len(parameters)):
                param_object = param_objects[param_names[i]]
                if param_object.annotation == ParamType.NUMBER or param_object.annotation == ParamType.POSITIVE_NUMBER:
                    try:
                        parameters[i] = float(parameters[i])
                    except:
                        await self.chat.send_message("the '{}' parameter must be numeric.".format(param_names[i]))
                        return True
                    if param_object.annotation == ParamType.POSITIVE_NUMBER and parameters[i] <= 0:
                        await self.chat.send_message("the '{}' parameter must be a positive number.".format(param_names[i]))
                        return True
                elif param_object.annotation == ParamType.MIXER_USER:
                    if parameters[i][:1] != "@":
                        await self.chat.send_message("the '{}' parameter must be a tagged user.".format(param_names[i]))
                        return True
                    try:
                        channel = await self.chat.api.get_channel(parameters[i][1:])
                        parameters[i] = channel.user
                    except:
                        await self.chat.send_message("the '{}' parameter must be a tagged user.".format(param_names[i]))
                        return True

            # NOTE:
            # the asyncio.ensure_future function is used rather than a standard await
            # since the executed command may contain async sleeping,
            # awaiting the call may freeze handling of incoming messages
            # https://docs.python.org/3/library/asyncio-future.html#asyncio.ensure_future
            coro = self.trigger_command(command, message, parameters)
            task = asyncio.ensure_future(coro)

            return True

        def __init__(self, chat, prefix):

            self.chat = chat
            self.prefix = prefix

            # initialize default commands
            for name, methods in DEFAULT_COMMANDS.items():
                for method in methods:
                    self.add_command(name, method)

        def __call__(self, method):
            self.add_command(method.__name__, method)

    # used to uniquely identify 'method' packets
    packet_id = 0

    # used to store references to functions (see __call__ and call_func)
    funcs = dict()
    callbacks = dict()

    # map events to functions
    event_map = {
        # ChatMessage -> handle_message (handled manually)
        "WelcomeEvent": "welcomed",
        "UserJoin": "user_joined",
        "UserLeave": "user_left",
        "PollStart": "poll_started",
        "PollEnd": "poll_end",
        "DeleteMessage": "message_deleted",
        "PurgeMessage": "messages_purged",
        "ClearMessages": "messages_cleared",
        "UserUpdate": "user_updated",
        "UserTimeout": "user_timed_out",
        "SkillAttribution": "handle_skill",
        "DeleteSkillAttribution": "skill_cancelled"
    }

    @classmethod
    async def create(cls, api, username_or_id, command_prefix = "!"):

        self = MixerChat()
        self.api = api
        self.channel = await self.api.get_channel(username_or_id)
        self.commands = self.ChatCommands(self, command_prefix)

        return self

    def __call__(self, method):
        if inspect.iscoroutinefunction(method):
            self.funcs[method.__name__] = method

    async def call_func(self, name, *args):

        # make sure the function exists
        # these are added via __call__ (@instance_name decorator)
        if not name in self.funcs: return

        # get a reference to the function
        func = self.funcs[name]

        # call the function
        await func(*args)

    async def send_method_packet(self, method, *args):
        """Sends a 'method' type packet to the Mixer chat websocket.

        Args:
            method (str): The method name.
            *args: List of arguments to pass to the server for this method.

        Returns:
            int: Unique packet identifier, used to register a callback.
        """
        packet = {
            "type": "method",
            "method": method,
            "arguments": list(args),
            "id": self.packet_id
        }
        await self.websocket.send_packet(packet)
        self.packet_id += 1
        return packet["id"]

    def register_method_callback(self, id, callback):
        """Creates a callback to handle replies to a method packet.

        Args:
            id (int): Unique packet ID returned by send_method_packet.
            callback (function): Callable to trigger when reply packet is received.
        """
        if inspect.iscoroutinefunction(callback):
            if not id in self.callbacks:
                self.callbacks[id] = callback

    async def start(self, oauth):
        """Initializes a websocket connection and starts handling commands.

        Args:
            oauth (MixerOAuth): Wrapper for access/refresh tokens.
        """

        url = "{}/chats/{}".format(self.api.API_URL, self.channel.id)
        response = requests.get(url, headers = oauth.header)
        chat_info = response.json() # https://pastebin.com/Z3RyUgBh

        # authentication callback (executed when w received reply for 'auth' method)
        async def auth_callback(data):
            if data["authenticated"]:
                await self.call_func("on_ready", oauth.username, oauth.user_id)

        # send auth packet upon connection and register auth_callback
        async def connected_callback():
            auth_packet_id = await self.send_method_packet("auth", self.channel.id, oauth.user_id, chat_info["authkey"])
            self.register_method_callback(auth_packet_id, auth_callback)

        # establish websocket connection and receive welcome packet
        self.websocket = MixerWS(chat_info["endpoints"][0])
        self.websocket.on_connected = connected_callback
        await self.websocket.connect()

        # infinite loop to handle future packets from server
        while True:

            # receive a packet from the server
            packet = await self.websocket.receive_packet()

            # handle 'event' packets from server
            if packet["type"] == "event":

                # custom handling for chat messages (commands and stuff)
                if packet["event"] == "ChatMessage":
                    message = MixerChatMessage(packet["data"])
                    message.chat = self
                    message.api = self.api
                    message.handled = await self.commands.handle(message)
                    await self.call_func("handle_message", message)
                    continue

                # call corresponding event handler
                if packet["event"] in self.event_map:
                    func_name = self.event_map[packet["event"]]
                    await self.call_func(func_name, packet["data"])

                continue

            # handle 'reply' packets from server
            if packet["type"] == "reply":

                # see if there's a reply callback for this packet
                callback = self.callbacks.pop(packet["id"], None)
                if callback is not None:

                    # invoke callback with data from reply packet
                    response = packet.get("data", packet)
                    await callback(response)

                continue

    async def send_message(self, message, user = None):
        """Send a message in the chat.

        Args:
            message (str): Message to send.
            user (str): Username to whisper to. Optional, will be sent in all chat if not provided.
        """
        if user is None:
            await self.send_method_packet("msg", message)
        else:
            await self.send_method_packet("whisper", user, message)

async def help_0(message):
    """Displays a list of commands that can be used in the chat."""

    chat = message.chat
    command_count = 0
    command_names = list()

    # build a list of command names/descriptions with params
    for name, commands in chat.commands.commands.items():

        variants = list()

        for command in commands:

            if command["description"] is None:
                continue

            command_count += 1
            variants.append(str(command["param_count"]))

        if len(variants) == 0:
            continue
        elif len(variants) == 1:
            command_names.append(name)
        else:
            param_counts = ", ".join(variants)
            command_names.append("{} ({})".format(name, param_counts))

    # delete original >help message
    await message.delete()

    # formate response messages
    message1 = "There are a total of {} commands: {}."
    message2 = "To see information about a specific command, use '{}help command_name'."
    message1 = message1.format(command_count, ", ".join(command_names))
    message2 = message2.format(chat.commands.prefix)

    # whisper formatted response messages to used
    await chat.send_message(message1, message.username)
    await asyncio.sleep(.5) # wait before sending second message :p
    await chat.send_message(message2, message.username)

async def help_1(message, name):
    """Provides a description of a specific command."""
    name = name.lower()
    return message.chat.commands.get_help(name)

async def help_2(message, name, parameter_count_or_name):
    """Provides a description of a command given a parameter count or parameter name."""
    chat = message.chat

    # if the second parameter is an int, assume they're specifying parameter count
    name = name.lower()
    try:
        return chat.commands.get_help(name, int(parameter_count_or_name))
    except ValueError: pass

    # fallback to get_help command if it doesnt exist
    if not name in chat.commands.commands:
        return chat.commands.get_help(name)

    # try to find a definition of the specified command with the given parameter name
    for command in chat.commands.commands[name]:
        if parameter_count_or_name in command["params"]:
            return chat.commands.get_help(name, command["param_count"])
    return "no variation of command '{}' has parameter named '{}'.".format(name, parameter_count_or_name)

DEFAULT_COMMANDS = {
    "help": [help_0, help_1, help_2]
}
